#include "trie.h"
//#include "list.h"

#include <limits.h>
#include <assert.h>

extern mode_t mode;

char r2c(const record *r, int i){ //i-th char of the string associated to a record
  switch(mode){
  case CNT:
    if(i<r->familyname_size)
      return r->familyname[i];
    else if(i==r->familyname_size)
      return '$';
    else if(i<r->familyname_size+1+r->firstname_size)
      return r->firstname[i-r->familyname_size-1];
    else if(i==r->familyname_size+1+r->firstname_size)
      return '$';
    else if(i<r->familyname_size+1+r->firstname_size+1+r->number_size)
      return r->number[i-r->familyname_size-1-r->firstname_size-1];
    else return '\0';
    break;
  case NTC:
    if(i<r->firstname_size)
      return r->firstname[i];
    else if(i==r->firstname_size)
      return '$';
    else if(i<r->firstname_size+1+r->number_size)
      return r->number[i-r->firstname_size-1];
    else if(i==r->firstname_size+1+r->number_size)
      return '$';
    else if(i<r->firstname_size+1+r->number_size+1+r->familyname_size)
      return r->familyname[i-r->firstname_size-1-r->number_size-1];
    else return '\0';
    break;
  case TCN:
    if(i<r->number_size)
      return r->number[i];
    else if(i==r->number_size)
      return '$';
    else if(i<r->number_size+1+r->familyname_size)
      return r->familyname[i-r->number_size-1];
    else if(i==r->number_size+1+r->familyname_size)
      return '$';
    else if(i<r->number_size+1+r->familyname_size+1+r->firstname_size)
      return r->firstname[i-r->number_size-1-r->familyname_size-1];
    else return '\0';
    break;
  default:
    assert(0
);
  }
}

trie *new_trie(){
  trie *t=calloc(1,sizeof(trie));
  t->len=0;
  t->child=NULL;
  t->parent=NULL;
  t->next=NULL;
  t->data=NULL;
  return t;
}

int is_leaf(const trie *t){
  return t->child==NULL;
}

int insert(trie *t, record *r){
  if(t->child==NULL && t->len==0){ //special case: empty trie
    t->data=r;
    t->len=INT_MAX;
    return 0;
  }
  
  trie *u=t;
  int i=0; //"iterator" in the record we are inserting
  while(1){ //XXXXXXXXx
    while(i<u->len){
      if(r2c(r,i)!=r2c(u->data,i)){ //insert "splitting" an existing node. labels are equal from index 0 to i-1.
	
	trie *t1=calloc(1,sizeof(trie)); //the "old" branch
	t1->len=u->len;
	t1->data=u->data;
	t1->parent=u;
	t1->child=u->child;
	
	trie *t2=calloc(1,sizeof(trie)); //the "new" branch
	t2->len=INT_MAX;
	t2->data=r;
	t2->parent=u;
	t2->child=NULL;
	
	u->len=i;
	
	//insert the new nodes, sorted
	if(r2c(u->data,i)<r2c(r,i)){
	  u->child=t1;
	  t1->next=t2;
	  t2->next=NULL; 
	}
	else{
	  u->child=t2;
	  t2->next=t1;
	  t1->next=NULL;
	}
	return 0;
      } else if (r2c(r,i)=='\0') return 1; //full match
      ++i;
    }
    //move to appropriate child
    trie **pc=&(u->child);
    while(*pc!=NULL && r2c((*pc)->data,i)<r2c(r,i))
      pc=&((*pc)->next);
    if(*pc==NULL || r2c((*pc)->data,i)>r2c(r,i)){//no exact match, add a new child to the list
      trie *t1=calloc(1,sizeof(trie));
      t1->len=INT_MAX;
      t1->child=NULL;
      t1->parent=u;
      t1->next=*pc;
      t1->data=r;
      *pc=t1;
      return 0;
    }
    u=*pc;
  } //while 1
}

record *lookup(const trie *t, const char *key){
  if(t->child==NULL && t->len==0) return NULL;
  const trie *u=t;
  int i=0;
  while(1){
    while(i<u->len){
      if(r2c(u->data,i)!=key[i]) return NULL; //no match
      if(key[i]=='\0') return u->data; //full match
      ++i;
    }
    trie *c=u->child;
    while(c!=NULL && r2c(c->data,i)<key[i]) c=c->next;
    if(c==NULL) return NULL;
    if(r2c(c->data,i) > key[i]) return NULL;
    u=c;
  }
}

void print_all(const trie *t){
  if(is_leaf(t)){
    if(t->data!=NULL)
      print_record(t->data);
    return;
  }
  else{
    trie *c=t->child;
    while(c!=NULL){
      print_all(c);
      c=c->next;
    }
    return;
  }
}

void print_prefix(const trie *t, const char *key){
  if(t->child==NULL && t->len==0) return; //empty trie
  if(key[0]=='\0') { //empty prefix
    print_all(t); 
    return;
  }
  const trie *u=t;
  int i=0;
  while(1){
    while(i<u->len){
      if(key[i]=='\0'){ //prefix fully matched
	print_all(u);
	return;
      }
      if(r2c(u->data,i)!=key[i]) return; //no match
      ++i;
    }

    //move to child
    trie *c=u->child;

    if(key[i]=='\0'){ //prefix ends on a child boundary: all children match
      while(c!=NULL){
	print_all(c);
	c=c->next;
      }
      return;
    }
    while(c!=NULL && r2c(c->data,i)<key[i]) c=c->next;
    if(c==NULL) return;
    if(r2c(c->data,i) > key[i]) return;
    u=c;
  }
}

record *delete(trie *t, record *r){
  if(t->child==NULL && t->len==0) return NULL;
  trie *u=t;
  int i=0;
  
  while(1){
    while(i<u->len){
      if(r2c(u->data,i)!=r2c(r,i)) return NULL; //no match
      if(r2c(r,i)=='\0'){//full match
	//ok, now we must delete trie u
	record *ret=u->data; //to be returned
	trie *par=u->parent;
	if(par==NULL){ //u was the trunk, so the trie is empty now
	  u->child=NULL;u->len=0;u->data=NULL;
	  //we must not delete u, since it /is/ the trie
	  return ret;
	}
	if(par->child==u){ //delete first child
	  par->child=u->next;
	}
	else{ //delete non-first child
	  trie *z=par->child;
	  while(z->next!=u) z=z->next;
	  z->next=u->next;
	}
	
	//is u->data (i.e. r) kept as "data" member somewhere around? If so, replace it with par->child->data
	trie *y=par;
	while(y!=NULL){
	  if(y->data==u->data)
	    y->data=par->child->data;
	  y=y->parent;
	}

	//par has become an unary node? if so, let's compact it
	if(par->child->next==NULL){
	  trie *x=par->child;
	  par->len=x->len;
	  par->data=x->data;
	  par->child=x->child;
	  trie *z=par->child;
	  //change parent of x's childs to par
	  while(z!=NULL){
	    z->parent=par;
	    z=z->next;
	  }
	  free(x);
	}

	free(u);
	return ret;
      }
      ++i;
    }
    trie *c=u->child; //move to childs
    while(c!=NULL && r2c(c->data,i)<r2c(r,i)) c=c->next;
    if(c==NULL) return NULL;
    if(r2c(c->data,i) > r2c(r,i)) return NULL;
    u=c;
  }

}
