#ifndef TRIE_H
#define TRIE_H

#include <stdlib.h>

#include "record.h"
//#include "list.h"

struct trie_tmp; //this is to make cross-includes work
typedef struct trie_tmp trie;

//compact trie using lists instead of arrays as a backend. This way we can deal with non-alphabetic entries and waste less space. Search time increases a bit, but we don't care: if the size of the alphabet is bounded, it is still constant-time

struct trie_tmp{
  int len; //total length of the "label", -1 if is a leaf
  trie *child; //points to the first child (the others are child->next, child->next->next...)
  trie *parent; //needed to update the key fields of parents on removal (the external string may become invalid)
  trie *next; //points to the next "brother", works as a list. Should always be sorted
  record *data;
  //if we need space, this can be compacted: one in child and data is always null, and we can decide which based on whether len==-1
};

trie *new_trie();

int is_leaf(const trie *t);

int insert(trie *t, record *r); //0=ok, 1=already existent

record *lookup(const trie *t, const char *key); //NULL=not found

void print_all(const trie *t);

void print_prefix(const trie *t, const char *key); //prints all elements matching the given prefix

record *delete(trie *t, record *r); //returns a pointer to the deleted record (which must still be free()'d) , NULL if not found

//modes for the three tries we will need: c=family name, n=first name, t=phone number
//e.g. CNT is for Bush$John W.$wh1teh0us3. $ used as a separator.
typedef enum{CNT, NTC, TCN} mode_t;

char r2c(const record *r,int i);

#endif //TRIE_H
